'' =================================================================================================
''
''   File....... hc-8+_dmx_remote_relays.spin
''   Purpose.... Basic DMX master controller for HC-8+ 
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2011-2015 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 20 NOV 2015
''
'' =================================================================================================

' This program will send relay control signals (0 for off, 255 for on) to the DMX stream starting
' at the address set on the DMX switch.
'
' Note: Do not use DMX.256 switch on Rev C or earlier boards; leave in the off position.


con

  PCB_REV = "C"


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                           ' use 5MHz crystal

  CLK_FREQ = _xinfreq * (_clkmode >> 6)                          ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                    ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                                ' ticks in 1us

  
con { io pins }

  RX1     = 31  { I }                                            ' serial / programming
  TX1     = 30  { O }                                            
                                                                 
  SDA     = 29  { I/O }                                          ' i2c / eeprom
  SCL     = 28  { I/O }                                          
                                                                 
  OUT_EN  = 27  { O }                                            ' outputs enable pin (low)     
                                                                 
  SIO     = 26  { I/O }                                          ' TTL serial io
                                                                 
  RX2     = 23  { I }                                            ' RS-485 port
  RX2_EN  = 22  { O }                                                 
  TX2_EN  = 21  { O }                                                 
  TX2     = 20  { O }                                                 
                                                                 
  R_LED   = 18  { O }                                            ' bi-color LED
  G_LED   = 17  { O }                                                 
                                                                 
  OUT7    = 15  { O }                                            ' outputs
  OUT6    = 14  { O }                                                 
  OUT5    = 13  { O }                                                 
  OUT4    = 12  { O }                                                 
  OUT3    = 11  { O }                                                 
  OUT2    = 10  { O }                                                 
  OUT1    =  9  { O }                                                 
  OUT0    =  8  { O }                                                 
                                                                 
  OPT_BR  =  7  { I/O }                                          ' options switches
  OPT_A1  =  6  { I/O }                                                  
  OPT_A0  =  5  { I/O }                                                  
  OPT_SM  =  4  { I/O }                                                  
                                                                 
  DMX_A8  =  3  { do not use with Rev C and earlier PCBs }       ' high-bit of DMX address
                                                                 
  DO_165  =  2  { I }                                            ' control lines for 74x165
  CLK_165 =  1  { O }                                                  
  LD_165  =  0  { O }                                                  


con

  #-1, ON, OFF                                                   ' digital control

  #0, BLK, RED, GRN, YEL                                         ' for r/g led
                                                                  
  #0, NOCON, RX, TX, RXTX                                        ' for rs-485

  #0, LSBFIRST, MSBFIRST                                        
                                                                  

obj

' main                                                           ' * master Spin cog

  time : "jm_time"                                               '   timing and delays
  io   : "jm_io_basic"                                           '   essential io control
  pwm  : "jm_pwm8"                                               ' * pwm control for LEDs
  dmx  : "jm_dmx_tx"                                             ' * dmx output to slaves
  
' * Uses cog when loaded
                                                             
                                                                  
var


pub main | t, addr, ch
                                                                  
  setup                                                          ' setup io and objects

  t := cnt
  repeat
    if (dmxaddr > 0)                                             ' okay to send values?
      repeat ch from 0 to 7                                      ' loop through inputs
        if (ttlpins & (1 << ch))                                 ' is TTL pin active?
          pwm.high(ch)                                           ' yes, on
          dmx.high(dmxaddr + ch)
        else
          pwm.low(ch)                                            ' no, off
          dmx.low(dmxaddr + ch)
          
    else
      pwm.set_all(0)
      dmx.clear
                                                                  
    waitcnt(t += constant(US_001 * (22_800 >> 2)))               ' 4x per standard dmx universe
                                                                           
                                                                  
pub setup                                                  
                                                                  
'' Configure HC-8+ IO and support objects

  time.start                                                     ' setup for timing/delays

  io.start(0, 0)                                                 ' clear pins (this cog)
                                                           
  outa[OUT7..OUT0] := %0000_0000                                 ' clear outputs
  dira[OUT7..OUT0] := %1111_1111                                 ' output mode                                  
  io.low(OUT_EN)                                                 ' enable output driver

  pwm.start(8, OUT0)                                             ' start outputs driver *

  dmx.start(TX2, TX2_EN)                                         ' start dmx output for slaves *

  bcog := cognew(background, @bcstack) + 1                       ' start background cog *


con

  { --------------------------------- }
  {  S U P P O R T   R O U T I N E S  } 
  { --------------------------------- }


var

  long  ttlpins                                                  ' updated by background cog
  long  dmxaddr
                                                                  

pub ttl_pin | btns

'' Reads TTL states for 1-of-8 input
'' -- returns -1 (no input)
'' -- returns 0 to 7 for input on IN0 to IN7
''    * only one input allowed
                                                                  
  if (bit_count(ttlpins) == 1)                                   ' only one button pressed?
    return bit_pos(ttlpins, 0)                                   '  yes, return button #
  else                                                            
    return -1                                                    '  none or bad input


pub ttl_inputs 

'' Return TTL inputs status
'' -- pins are scanned by background cog

  return ttlpins


pub dmx_address 

'' Return DMX address switch setting
'' -- switches are scanned by background cog
'' -- Note: For Rev C and lower boards, do not use DMX 256 switch -- leave it OFF

  return dmxaddr


pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                  ' force to input mode
                                                                  
  return !ina[OPT_BR..OPT_SM] & $0F                              ' return state (1 = on)
                                                                  
                                                                  
pub set_rg_led(state)                                              
                                                                  
'' Sets R/G LED                                                   
'' -- direct control only                                         
                                                                  
  case state                                                     ' set color to state
    OFF : outa[R_LED..G_LED] := %00                             
    RED : outa[R_LED..G_LED] := %10                             
    GRN : outa[R_LED..G_LED] := %01                             
    YEL : outa[R_LED..G_LED] := %00                              ' yellow requires active driver

  dira[R_LED..G_LED] := %11                                      ' make outputs
   
                                                     
pub set_rs485(state)                                              
                                                                  
'' Sets RS-485 RX and TX enable inputs                            
                                                                  
  case state                                                      
    NOCON : outa[RX2_EN..TX2_EN] := %10                           
    RX    : outa[RX2_EN..TX2_EN] := %00                           
    TX    : outa[RX2_EN..TX2_EN] := %11                           
    RXTX  : outa[RX2_EN..TX2_EN] := %01                           
                                                                  
  dira[RX2_EN..TX2_EN] := %11                                     
                                                                  
                                                                  
pub bit_count(value) | bc                                         
                                                                  
'' Returns # of bits set (1) in value                             
                                                                  
  bc := 0                                                        ' clear count
  repeat 32                                                      ' test all bits
    bc += (value & %1)                                           ' add bit value to count
    value >>= 1                                                  ' next bit
                                                                  
  return bc                                                       
                                                                  
                                                                  
pub bit_pos(value, mode)                                          
                                                                  
'' Returns position of 1st bit set (0..31)                        
'' -- mode 0 to scan from lsb, mode 1 to scan from msb            
'' -- -1 = no bits set                                            
                                                                  
  if (value == 0)                                                ' if no bits                                   
    return -1                                                    '  return -1
                                                                  
  else                                                            
    if (mode == LSBFIRST)                                        ' check from LSB
      value ><= 32                                               '  flip for >|
      return (32 - >|value)                                       
    else                                                          
      return (>|value - 1)                                        
                                                                  
                                                                  
pub bit_val(value, pos)                                           
                                                            
'' Returns bit (0..1) from pos (0..31) in value                   
                                                                  
  if ((pos => 0) and (pos =< 31))                                 
    return (value >> pos) & 1                                     
  else                                                            
    return 0


con

  { ----------------------------- }
  {  B A C K G R O U N D   C O G  }
  {  - global milliseconds        }
  {  - TTL/DMX inputs scanning    }
  { ----------------------------- }
  

var  

  long  bcog                                                     ' background cog #
  long  bcstack[64]                                              ' stack for background cog

  long  millis                                                   ' global milliseconds register

  
pri background | t                                               ' launch with cognew()

  io.low(R_LED)                                                  ' setup R/G LED pins
  io.low(G_LED)

  io.high(LD_165)                                                ' setup x165 io pins
  io.low(CLK_165)
  io.input(DO_165)

  millis := 0

  t := cnt                                                       ' sync loop timer
  repeat
    waitcnt(t += MS_001)                                         ' run loop every millisecond
    ++millis
    scan_inputs                                                  ' re-scan TTL & DMX inputs

          
pri scan_inputs | tempin

'' Scan TTL and DMX address inputs

  outa[LD_165] := 0                                              ' blip Shift/Load line    
  outa[LD_165] := 1

  tempin := 0
       
  repeat 16                                                      ' read n bits         
    tempin := (tempin << 1) | ina[DO_165]                        ' get new bit             
    outa[CLK_165] := 1                                           ' blip clock              
    outa[CLK_165] := 0

  ttlpins := tempin & $FF                                        ' update global vars
  dmxaddr := tempin >> 8                                         
                                                         

dat { license }

{{

  Copyright (c) 2011-2015 Jon McPhalen     

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      